#ifndef _AST_HPP
#  define _AST_HPP

/* Automatically generated by AST.sh -- do not edit */

#  include "IR.hpp"
#  include <string>
#  include <vector>
#  include <map>

namespace AST {
  using IR::Node;
  using IR::XMListream;
  using IR::XMLostream;
  using std::string;
  using std::vector;
  using std::map;

  class Visitor;

  class ASTNode;
  class Symbol;
  class ModuleSymbol;
  class SignalSymbol;
  class BuiltinSignalSymbol;
  class TrapSymbol;
  class VariableSymbol;
  class ConstantSymbol;
  class BuiltinConstantSymbol;
  class TypeSymbol;
  class BuiltinTypeSymbol;
  class FunctionSymbol;
  class BuiltinFunctionSymbol;
  class ProcedureSymbol;
  class TaskSymbol;
  class SymbolTable;
  class Expression;
  class Literal;
  class LoadVariableExpression;
  class LoadSignalExpression;
  class LoadSignalValueExpression;
  class LoadTrapExpression;
  class LoadTrapValueExpression;
  class UnaryOp;
  class BinaryOp;
  class FunctionCall;
  class Delay;
  class CheckCounter;
  class Module;
  class InputRelation;
  class Exclusion;
  class Implication;
  class Modules;
  class Statement;
  class BodyStatement;
  class PredicatedStatement;
  class CaseStatement;
  class StatementList;
  class ParallelStatementList;
  class Nothing;
  class Pause;
  class Halt;
  class StartCounter;
  class Emit;
  class Sustain;
  class Assignment;
  class ProcedureCall;
  class Present;
  class If;
  class Loop;
  class Repeat;
  class Abort;
  class Await;
  class LoopEach;
  class Every;
  class Suspend;
  class DoWatching;
  class DoUpto;
  class TaskCall;
  class Exec;
  class ScopeStatement;
  class Trap;
  class Exit;
  class Signal;
  class Var;
  class Renaming;
  class TypeRenaming;
  class ConstantRenaming;
  class FunctionRenaming;
  class ProcedureRenaming;
  class SignalRenaming;
  class Run;
  class IfThenElse;
  class Goto;
  class Label;
  class Catch;
  class Handler;
  class Try;
  class Break;
  class Continuable;
  class Resume;
  class Continue;
  class Thread;
  class Parallel;
  class GRCNode;
  class STNode;
  class STexcl;
  class STpar;
  class STref;
  class STleaf;
  class TopGRC;
  class Nop;
  class RecT1;
  class DefineSignal;
  class UndefineSignal;
  class Switch;
  class Test;
  class Fork;
  class Sync;
  class Terminate;
  class Action;
  class Enter;
  class Leave;

  union Status {
    int i;
    ASTNode *n;
    Status() {}
    Status(int ii) : i(ii) {}
    Status(ASTNode *nn) : n(nn) {}
  };



  class ASTNode : public Node {
    public:
    void read(XMListream &);
    void write(XMLostream &) const;
    ASTNode() {}



  virtual Status welcome(Visitor&) = 0;
  };


  class Symbol : public ASTNode {
    public:
    void read(XMListream &);
    void write(XMLostream &) const;
    Symbol() {}

  string name;

  Symbol(string s) : name(s) {}
  };


  class ModuleSymbol : public Symbol {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    ModuleSymbol() {}

  Module *module;

  ModuleSymbol(string s) : Symbol(s), module(0) {}
  };


  class SignalSymbol : public Symbol {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    SignalSymbol() {}

  TypeSymbol *type;
  string direction; // input, output, inputoutput, sensor, return, local
  FunctionSymbol *combine; // combining function, if any
  Expression *initializer;
  VariableSymbol *presence;
  VariableSymbol *value;  

  SignalSymbol(string n, TypeSymbol *t, string d, FunctionSymbol *f,
               Expression *e, VariableSymbol *p, VariableSymbol *v)
    : Symbol(n), type(t), direction(d), combine(f), initializer(e),
      presence(p), value(v) {}
  };


  class BuiltinSignalSymbol : public SignalSymbol {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    BuiltinSignalSymbol() {}


  BuiltinSignalSymbol(string n, TypeSymbol *t, string d, FunctionSymbol *f,
                      VariableSymbol *p, VariableSymbol *v)
    : SignalSymbol(n, t, d, f, NULL, p, v) {}
  };


  class TrapSymbol : public Symbol {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    TrapSymbol() {}

  TypeSymbol *type;
  Expression *initializer;
  int code;
  VariableSymbol *presence;
  VariableSymbol *value;

  TrapSymbol(string s, TypeSymbol *t, Expression *e, VariableSymbol *p,
             VariableSymbol *v)
    : Symbol(s), type(t), initializer(e), code(0), presence(p), value(v) {}
  };


  class VariableSymbol : public Symbol {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    VariableSymbol() {}

  TypeSymbol *type;
  Expression *initializer;

  VariableSymbol(string n, TypeSymbol *t, Expression *e)
    : Symbol(n), type(t), initializer(e) {}
  };


  class ConstantSymbol : public VariableSymbol {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    ConstantSymbol() {}


  ConstantSymbol(string n, TypeSymbol *t, Expression *i)
     : VariableSymbol(n, t, i) {}
  };


  class BuiltinConstantSymbol : public ConstantSymbol {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    BuiltinConstantSymbol() {}


  BuiltinConstantSymbol(string n, TypeSymbol *t, Expression *i)
     : ConstantSymbol(n, t, i) {}
  };


  class TypeSymbol : public Symbol {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    TypeSymbol() {}


   TypeSymbol(string s) : Symbol(s) {}
  };


  class BuiltinTypeSymbol : public TypeSymbol {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    BuiltinTypeSymbol() {}


   BuiltinTypeSymbol(string s) : TypeSymbol(s) {}
  };


  class FunctionSymbol : public TypeSymbol {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    FunctionSymbol() {}

  vector<TypeSymbol*> arguments;
  TypeSymbol *result;

  FunctionSymbol(string s) : TypeSymbol(s), result(NULL) {}
  };


  class BuiltinFunctionSymbol : public FunctionSymbol {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    BuiltinFunctionSymbol() {}


  BuiltinFunctionSymbol(string s) : FunctionSymbol(s) {}
  };


  class ProcedureSymbol : public TypeSymbol {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    ProcedureSymbol() {}

  vector<TypeSymbol*> reference_arguments;
  vector<TypeSymbol*> value_arguments;

  ProcedureSymbol(string s) : TypeSymbol(s) {}
  };


  class TaskSymbol : public ProcedureSymbol {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    TaskSymbol() {}


  TaskSymbol(string s) : ProcedureSymbol(s) {}
  };


  class SymbolTable : public ASTNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    
  SymbolTable *parent;
  typedef map<string, Symbol*> stmap;
  stmap symbols;

  SymbolTable() : parent(NULL) {}

  class iterator {
    stmap::iterator i;
  public:
    iterator(stmap::iterator ii) : i(ii) {}
    void operator ++(int) { i++; } // int argument denotes postfix
    void operator ++() { ++i; } // int argument denotes postfix
    bool operator !=(const iterator &ii) { return i != ii.i; }
    Symbol *operator *() { return (*i).second; }
  };

  iterator begin() { return iterator(symbols.begin()); }
  iterator end() { return iterator(symbols.end()); }

  bool local_contains(const string) const;
  bool contains(const string) const;
  void enter(Symbol *);
  Symbol* get(const string);
  };


  class Expression : public ASTNode {
    public:
    void read(XMListream &);
    void write(XMLostream &) const;
    Expression() {}

  TypeSymbol *type;

  Expression(TypeSymbol *t) : type(t) {}
  };


  class Literal : public Expression {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Literal() {}

  string value;

  Literal(string v, TypeSymbol *t) : Expression(t), value(v) {}
  };


  class LoadVariableExpression : public Expression {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    LoadVariableExpression() {}

  VariableSymbol *variable;

  LoadVariableExpression(VariableSymbol *v)
    : Expression(v->type), variable(v) {}
  };


  class LoadSignalExpression : public Expression {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    LoadSignalExpression() {}

  SignalSymbol *signal;

  LoadSignalExpression(SignalSymbol *s)
    : Expression(s->type), signal(s) {}
  };


  class LoadSignalValueExpression : public LoadSignalExpression {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    LoadSignalValueExpression() {}

  
  LoadSignalValueExpression(SignalSymbol *s) : LoadSignalExpression(s) {}
  };


  class LoadTrapExpression : public Expression {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    LoadTrapExpression() {}

  TrapSymbol *trap;

  LoadTrapExpression(TrapSymbol *t)
   : Expression(t->type), trap(t) {}
  };


  class LoadTrapValueExpression : public LoadTrapExpression {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    LoadTrapValueExpression() {}

  
  LoadTrapValueExpression(TrapSymbol *s) : LoadTrapExpression(s) {}
  };


  class UnaryOp : public Expression {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    UnaryOp() {}

  string op;
  Expression *source;

  UnaryOp(TypeSymbol *t, string s, Expression *e)
    : Expression(t), op(s), source(e) {}
  };


  class BinaryOp : public Expression {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    BinaryOp() {}

  string op;
  Expression *source1;
  Expression *source2;

  BinaryOp(TypeSymbol *t, string s, Expression *e1, Expression *e2)
    : Expression(t), op(s), source1(e1), source2(e2) {}
  };


  class FunctionCall : public Expression {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    FunctionCall() {}

  FunctionSymbol *callee;
  vector<Expression*> arguments;

  FunctionCall(FunctionSymbol *s)
    : Expression(s->result), callee(s) {}
  };


  class Delay : public Expression {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Delay() {}

  Expression *predicate;
  Expression *count;
  VariableSymbol *counter;
  bool is_immediate;

  Delay(TypeSymbol *t, Expression *e1, Expression *e2, VariableSymbol *v,
         bool i) : Expression(t), predicate(e1), count(e2), counter(v),
                   is_immediate(i) {}
  };


  class CheckCounter : public Expression {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    CheckCounter() {}

  VariableSymbol *counter;

  CheckCounter(TypeSymbol *t, VariableSymbol *c): Expression(t), counter(c) {}
  };


  class Module : public ASTNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    
  ModuleSymbol *symbol;
  SymbolTable *types;
  SymbolTable *constants;
  SymbolTable *functions;
  SymbolTable *procedures;
  SymbolTable *tasks;
  SymbolTable *signals;
  SymbolTable *variables;
  vector<InputRelation*> relations;
  ASTNode *body;
  int max_code;
  
  Module() : max_code(0) {}
  Module(ModuleSymbol *);
  ~Module();    
  };


  class InputRelation : public ASTNode {
    public:
    void read(XMListream &);
    void write(XMLostream &) const;
    InputRelation() {}


  };


  class Exclusion : public InputRelation {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Exclusion() {}

  vector<SignalSymbol *> signals;
  };


  class Implication : public InputRelation {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Implication() {}

  SignalSymbol *predicate;
  SignalSymbol *implication;

  Implication(SignalSymbol *ss1, SignalSymbol*ss2)
    : predicate(ss1), implication(ss2) {}
  };


  class Modules : public ASTNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Modules() {}

  SymbolTable module_symbols;
  vector<Module*> modules;

  void add(Module*);
  };


  class Statement : public ASTNode {
    public:
    void read(XMListream &);
    void write(XMLostream &) const;
    Statement() {}


  };


  class BodyStatement : public Statement {
    public:
    void read(XMListream &);
    void write(XMLostream &) const;
    BodyStatement() {}

  Statement *body;

  BodyStatement(Statement *s) : body(s) {}
  };


  class PredicatedStatement : public BodyStatement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    PredicatedStatement() {}

  Expression *predicate;

  PredicatedStatement(Statement *s, Expression *e)
    : BodyStatement(s), predicate(e) {}
  };


  class CaseStatement : public Statement {
    public:
    void read(XMListream &);
    void write(XMLostream &) const;
    
  vector<PredicatedStatement *> cases;
  Statement *default_stmt;

  CaseStatement() : default_stmt(0) {}
  PredicatedStatement *newCase(Statement *s, Expression *e) {
    PredicatedStatement *ps = new PredicatedStatement(s, e);
    cases.push_back(ps);
    return ps;
  }
  };


  class StatementList : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    StatementList() {}

  vector<Statement *> statements;

  StatementList& operator <<(Statement *s) {
    assert(s);
    statements.push_back(s);
    return *this;
  }
  };


  class ParallelStatementList : public Statement  {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    ParallelStatementList() {}

  vector<Statement *> threads;
  };


  class Nothing : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Nothing() {}


  };


  class Pause : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Pause() {}


  };


  class Halt : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Halt() {}


  };


  class StartCounter : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    StartCounter() {}

  Expression *predicate;
  Expression *startvalue;
  VariableSymbol *counter;
  STNode *st;

  StartCounter(Expression *p, Expression *sv, VariableSymbol *v): predicate(p), startvalue(sv), counter(v) {}
  };


  class Emit : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Emit() {}

  SignalSymbol *signal;
  Expression *value;

  Emit(SignalSymbol *s, Expression *e) : signal(s), value(e) {}
  };


  class Sustain : public Emit {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Sustain() {}


  Sustain(SignalSymbol *s, Expression *e) : Emit(s, e) {}
  };


  class Assignment : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Assignment() {}

  VariableSymbol *variable;
  Expression *value;

  Assignment(VariableSymbol *v, Expression *e) : variable(v), value(e) {}
  };


  class ProcedureCall : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    ProcedureCall() {}

  ProcedureSymbol *procedure;
  vector<VariableSymbol*> reference_args;
  vector<Expression*> value_args;

  ProcedureCall(ProcedureSymbol *ps) : procedure(ps) {}
  };


  class Present : public CaseStatement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Present() {}


  };


  class If : public CaseStatement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    If() {}


  };


  class Loop : public BodyStatement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Loop() {}


  Loop(Statement *s) : BodyStatement(s) {}
  };


  class Repeat : public Loop {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Repeat() {}

  Expression *count;
  bool is_positive;

  Repeat(Statement *s, Expression *e, bool p)
    : Loop(s), count(e), is_positive(p) {}
  };


  class Abort : public CaseStatement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Abort() {}

  Statement *body;
  bool is_weak;
  int code;

  Abort(Statement *s, bool i) : body(s), is_weak(i), code(0) {}
  Abort(Statement *s, Expression *e, Statement *s1)
    : body(s), is_weak(false), code(0) {
    newCase(s1, e);
  }
  };


  class Await : public CaseStatement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Await() {}


  };


  class LoopEach : public PredicatedStatement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    LoopEach() {}


  LoopEach(Statement *s, Expression *e) : PredicatedStatement(s, e) {}
  };


  class Every : public PredicatedStatement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Every() {}


  Every(Statement *s, Expression *e) : PredicatedStatement(s, e) {}
  };


  class Suspend : public PredicatedStatement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Suspend() {}


  Suspend(Statement *s, Expression *e) : PredicatedStatement(s, e) {}
  };


  class DoWatching : public PredicatedStatement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    DoWatching() {}

  Statement *timeout;

  DoWatching(Statement *s1, Expression *e, Statement *s2)
    : PredicatedStatement(s1, e), timeout(s2) {}
  };


  class DoUpto : public PredicatedStatement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    DoUpto() {}


  DoUpto(Statement *s, Expression *e) : PredicatedStatement(s, e) {}
  };


  class TaskCall : public ProcedureCall {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    TaskCall() {}

  SignalSymbol *signal;
  Statement *body;

  TaskCall(TaskSymbol *ts) : ProcedureCall(ts), signal(0), body(0) {}
  };


  class Exec : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Exec() {}

  vector <TaskCall *> calls;
  };


  class ScopeStatement : public BodyStatement {
    public:
    void read(XMListream &);
    void write(XMLostream &) const;
    ScopeStatement() {}

   SymbolTable *symbols;
  };


  class Trap : public ScopeStatement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Trap() {}

  vector<PredicatedStatement *> handlers;

  PredicatedStatement* newHandler(Expression *e, Statement *s) {
    PredicatedStatement *ps = new PredicatedStatement(s, e);
    handlers.push_back(ps);
    return ps;
  }
  };


  class Exit : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Exit() {}

  TrapSymbol *trap;
  Expression *value;

  Exit(TrapSymbol *t, Expression *e) : trap(t), value(e) {}
  };


  class Signal : public ScopeStatement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Signal() {}


  };


  class Var : public ScopeStatement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Var() {}


  };


  class Renaming : public ASTNode {
    public:
    void read(XMListream &);
    void write(XMLostream &) const;
    Renaming() {}

  string old_name;

Renaming(string s) : old_name(s) {}
  };


  class TypeRenaming : public Renaming {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    TypeRenaming() {}

  TypeSymbol *new_type;

  TypeRenaming(string s, TypeSymbol *t) : Renaming(s), new_type(t) {}
  };


  class ConstantRenaming : public Renaming {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    ConstantRenaming() {}

  Expression *new_value;

  ConstantRenaming(string s, Expression *e) : Renaming(s), new_value(e) {}
  };


  class FunctionRenaming : public Renaming {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    FunctionRenaming() {}

  FunctionSymbol *new_func;

  FunctionRenaming(string s, FunctionSymbol *f) : Renaming(s), new_func(f) {}
  };


  class ProcedureRenaming : public Renaming {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    ProcedureRenaming() {}

  ProcedureSymbol *new_proc;

  ProcedureRenaming(string s, ProcedureSymbol *p)
    : Renaming(s), new_proc(p) {}
  };


  class SignalRenaming : public Renaming {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    SignalRenaming() {}

  SignalSymbol *new_sig;

  SignalRenaming(string s, SignalSymbol *ss) : Renaming(s), new_sig(ss) {}
  };


  class Run : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Run() {}

  string old_name;
  string new_name;
  vector<TypeRenaming *> types;
  vector<ConstantRenaming *> constants;
  vector<FunctionRenaming *> functions;
  vector<ProcedureRenaming *> procedures;
  vector<ProcedureRenaming *> tasks;
  vector<SignalRenaming *> signals;

  Run(string s) : old_name(s), new_name(s) {}
  };


  class IfThenElse : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    IfThenElse() {}

  Expression *predicate;
  Statement *then_part;
  Statement *else_part;

  IfThenElse(Expression *e) : predicate(e) , then_part(0), else_part(0) {}
  IfThenElse(Expression *e, Statement *s1, Statement *s2)
    : predicate(e) , then_part(s1), else_part(s2) {}
  };


  class Goto : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Goto() {}

  Label *target;

  Goto(Label *s) : target(s) {}
  };


  class Label : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Label() {}

  string name;

  Label(string s) : name(s) {}
  };


  class Catch : public BodyStatement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Catch() {}

  int code;
  Label *label;

  Catch(Statement *s, int c, Label *l) : BodyStatement(s), code(c), label(l) {}
  };


  class Handler : public Statement {
    public:
    void read(XMListream &);
    void write(XMLostream &) const;
    Handler() {}

  vector<Catch*> catches;
  Label *catch0;

  Handler(Label *c0) : catch0(c0) {}
  void newCatch(Statement *s, int c, Label *l) {
    catches.push_back(new Catch(s, c, l));
  }
  };


  class Try : public Handler {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Try() {}

  Statement *body;

  Try(Label *c0, Statement *s) : Handler(c0), body(s) {}
  };


  class Break : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Break() {}

  int level;
  Label *continue_target;
  Statement *set_state;

  Break(int l) : level(l), continue_target(0), set_state(0) {}
  };


  class Continuable : public Handler {
    public:
    void read(XMListream &);
    void write(XMLostream &) const;
    Continuable() {}

  Label *continue_label;

  Continuable(Label *c0, Label *cont) : Handler(c0), continue_label(cont) {}
  };


  class Resume : public Continuable {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Resume() {}

  Statement *body;
  Statement *branch;

  Resume(Label *c0, Label *cont, Statement *s)
    : Continuable(c0, cont), body(s), branch(0) {}
  };


  class Continue : public Statement {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Continue() {}


  };


  class Thread : public Resume {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Thread() {}

  VariableSymbol *exit_level;

  Thread(Label *c0, Label *cont, Statement *b, VariableSymbol *el)
     : Resume(c0, cont, b), exit_level(el) {}
  };


  class Parallel : public Continuable {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Parallel() {}

  vector<Thread*> threads;
  Statement *synchronize;

  Parallel(Label *c0, Label *cont)
    : Continuable(c0, cont), synchronize(0) {}
  void newThread(Label *c0, Label *cont, Statement *b, VariableSymbol *el) {
    threads.push_back(new Thread(c0, cont, b, el));
  }
  };


  class GRCNode : public ASTNode {
    public:
    void read(XMListream &);
    void write(XMLostream &) const;
    GRCNode() {}

  vector<GRCNode*> predecessors;
  vector<GRCNode*> successors;

  virtual Status welcome(Visitor&) = 0;

  GRCNode& operator >>(GRCNode*);
  };


  class STNode : public GRCNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    STNode() {}


  };


  class STexcl : public STNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    STexcl() {}


  };


  class STpar : public STNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    STpar() {}


  };


  class STref : public STNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    
  int type;

  STref(): type(0) {}

  int isabort() { return type == 1;}
  void setabort() { type = 1;}
  int issuspend() { return type == 2;}
  void setsuspend() { type = 2;}
  };


  class STleaf : public STNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    
  int type;

  STleaf(): type(0) {}
 
  int isfinal() { return type == 1;}
  void setfinal() { type = 1;}
  };


  class TopGRC : public GRCNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    TopGRC() {}


  };


  class Nop : public GRCNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    
  int type;
  int code;

  Nop(): type(0), code(0) {}

  int isflowin() { return type == 1;}
  void setflowin() { type = 1;}
  // a shorcut Nop gives up flow to child 0
  int isshortcut() { return type == 2;}
  void setshortcut() { type = 2;}
  };


  class RecT1 : public GRCNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    RecT1() {}


  };


  class DefineSignal : public GRCNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    DefineSignal() {}

  SymbolTable *signals;

  DefineSignal(Signal &s) : signals(s.symbols){}
  };


  class UndefineSignal : public GRCNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    UndefineSignal() {}

  SymbolTable *signals;

  UndefineSignal(Signal &s): signals(s.symbols) {}
  };


  class Switch : public GRCNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Switch() {}

  STNode *st;  
  };


  class Test : public GRCNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Test() {}

  Expression *predicate;
  STNode *st;
  
  Test(Expression *e) : predicate(e) {}
  };


  class Fork : public GRCNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Fork() {}


  };


  class Sync : public GRCNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    
  STNode *st;  
  int depth;
 
  Sync() : depth(0) {}

  void setdepth() { depth = 1; }
  int isdepth() { return depth == 1; }
  };


  class Terminate : public GRCNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Terminate() {}

  int code;

  Terminate(int c) : code(c) {}
  };


  class Action : public GRCNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Action() {}

  Statement *body;

  Action(Statement *s) : body(s) {}
  };


  class Enter : public GRCNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Enter() {}

  STNode *st;
  };


  class Leave : public GRCNode {
    
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);
    void read(XMListream &);
    void write(XMLostream &) const;
    Leave() {}

  STNode *st;
  };


  class Visitor {
  public:

  virtual Status visit(ModuleSymbol& n) { assert(0); return Status(); }
  virtual Status visit(SignalSymbol& n) { assert(0); return Status(); }
  virtual Status visit(BuiltinSignalSymbol& n) { assert(0); return Status(); }
  virtual Status visit(TrapSymbol& n) { assert(0); return Status(); }
  virtual Status visit(VariableSymbol& n) { assert(0); return Status(); }
  virtual Status visit(ConstantSymbol& n) { assert(0); return Status(); }
  virtual Status visit(BuiltinConstantSymbol& n) { assert(0); return Status(); }
  virtual Status visit(TypeSymbol& n) { assert(0); return Status(); }
  virtual Status visit(BuiltinTypeSymbol& n) { assert(0); return Status(); }
  virtual Status visit(FunctionSymbol& n) { assert(0); return Status(); }
  virtual Status visit(BuiltinFunctionSymbol& n) { assert(0); return Status(); }
  virtual Status visit(ProcedureSymbol& n) { assert(0); return Status(); }
  virtual Status visit(TaskSymbol& n) { assert(0); return Status(); }
  virtual Status visit(SymbolTable& n) { assert(0); return Status(); }
  virtual Status visit(Literal& n) { assert(0); return Status(); }
  virtual Status visit(LoadVariableExpression& n) { assert(0); return Status(); }
  virtual Status visit(LoadSignalExpression& n) { assert(0); return Status(); }
  virtual Status visit(LoadSignalValueExpression& n) { assert(0); return Status(); }
  virtual Status visit(LoadTrapExpression& n) { assert(0); return Status(); }
  virtual Status visit(LoadTrapValueExpression& n) { assert(0); return Status(); }
  virtual Status visit(UnaryOp& n) { assert(0); return Status(); }
  virtual Status visit(BinaryOp& n) { assert(0); return Status(); }
  virtual Status visit(FunctionCall& n) { assert(0); return Status(); }
  virtual Status visit(Delay& n) { assert(0); return Status(); }
  virtual Status visit(CheckCounter& n) { assert(0); return Status(); }
  virtual Status visit(Module& n) { assert(0); return Status(); }
  virtual Status visit(Exclusion& n) { assert(0); return Status(); }
  virtual Status visit(Implication& n) { assert(0); return Status(); }
  virtual Status visit(Modules& n) { assert(0); return Status(); }
  virtual Status visit(PredicatedStatement& n) { assert(0); return Status(); }
  virtual Status visit(StatementList& n) { assert(0); return Status(); }
  virtual Status visit(ParallelStatementList& n) { assert(0); return Status(); }
  virtual Status visit(Nothing& n) { assert(0); return Status(); }
  virtual Status visit(Pause& n) { assert(0); return Status(); }
  virtual Status visit(Halt& n) { assert(0); return Status(); }
  virtual Status visit(StartCounter& n) { assert(0); return Status(); }
  virtual Status visit(Emit& n) { assert(0); return Status(); }
  virtual Status visit(Sustain& n) { assert(0); return Status(); }
  virtual Status visit(Assignment& n) { assert(0); return Status(); }
  virtual Status visit(ProcedureCall& n) { assert(0); return Status(); }
  virtual Status visit(Present& n) { assert(0); return Status(); }
  virtual Status visit(If& n) { assert(0); return Status(); }
  virtual Status visit(Loop& n) { assert(0); return Status(); }
  virtual Status visit(Repeat& n) { assert(0); return Status(); }
  virtual Status visit(Abort& n) { assert(0); return Status(); }
  virtual Status visit(Await& n) { assert(0); return Status(); }
  virtual Status visit(LoopEach& n) { assert(0); return Status(); }
  virtual Status visit(Every& n) { assert(0); return Status(); }
  virtual Status visit(Suspend& n) { assert(0); return Status(); }
  virtual Status visit(DoWatching& n) { assert(0); return Status(); }
  virtual Status visit(DoUpto& n) { assert(0); return Status(); }
  virtual Status visit(TaskCall& n) { assert(0); return Status(); }
  virtual Status visit(Exec& n) { assert(0); return Status(); }
  virtual Status visit(Trap& n) { assert(0); return Status(); }
  virtual Status visit(Exit& n) { assert(0); return Status(); }
  virtual Status visit(Signal& n) { assert(0); return Status(); }
  virtual Status visit(Var& n) { assert(0); return Status(); }
  virtual Status visit(TypeRenaming& n) { assert(0); return Status(); }
  virtual Status visit(ConstantRenaming& n) { assert(0); return Status(); }
  virtual Status visit(FunctionRenaming& n) { assert(0); return Status(); }
  virtual Status visit(ProcedureRenaming& n) { assert(0); return Status(); }
  virtual Status visit(SignalRenaming& n) { assert(0); return Status(); }
  virtual Status visit(Run& n) { assert(0); return Status(); }
  virtual Status visit(IfThenElse& n) { assert(0); return Status(); }
  virtual Status visit(Goto& n) { assert(0); return Status(); }
  virtual Status visit(Label& n) { assert(0); return Status(); }
  virtual Status visit(Catch& n) { assert(0); return Status(); }
  virtual Status visit(Try& n) { assert(0); return Status(); }
  virtual Status visit(Break& n) { assert(0); return Status(); }
  virtual Status visit(Resume& n) { assert(0); return Status(); }
  virtual Status visit(Continue& n) { assert(0); return Status(); }
  virtual Status visit(Thread& n) { assert(0); return Status(); }
  virtual Status visit(Parallel& n) { assert(0); return Status(); }
  virtual Status visit(STNode& n) { assert(0); return Status(); }
  virtual Status visit(STexcl& n) { assert(0); return Status(); }
  virtual Status visit(STpar& n) { assert(0); return Status(); }
  virtual Status visit(STref& n) { assert(0); return Status(); }
  virtual Status visit(STleaf& n) { assert(0); return Status(); }
  virtual Status visit(TopGRC& n) { assert(0); return Status(); }
  virtual Status visit(Nop& n) { assert(0); return Status(); }
  virtual Status visit(RecT1& n) { assert(0); return Status(); }
  virtual Status visit(DefineSignal& n) { assert(0); return Status(); }
  virtual Status visit(UndefineSignal& n) { assert(0); return Status(); }
  virtual Status visit(Switch& n) { assert(0); return Status(); }
  virtual Status visit(Test& n) { assert(0); return Status(); }
  virtual Status visit(Fork& n) { assert(0); return Status(); }
  virtual Status visit(Sync& n) { assert(0); return Status(); }
  virtual Status visit(Terminate& n) { assert(0); return Status(); }
  virtual Status visit(Action& n) { assert(0); return Status(); }
  virtual Status visit(Enter& n) { assert(0); return Status(); }
  virtual Status visit(Leave& n) { assert(0); return Status(); }
  };

}

#endif

