#line 1124 "ASTGRC.nw"
#ifndef _ASTGRC_HPP
#  define _ASTGRC_HPP

#  include "AST.hpp"
#  include <assert.h>
#  include <stack>

namespace ASTGRC {
  using namespace IR;
  using namespace AST;

  class GrcSynth;

  
#line 24 "ASTGRC.nw"
struct Context {
  int size;
  std::stack<GRCNode**> continuations;

  Context(int max_code) : size(max_code) {
    continuations.push(new GRCNode*[size]);
    for (int i = 0 ; i < size ; i++ ) continuations.top()[i] = 0;
  }
  ~Context() {}

  void push(Context &c) {
    GRCNode **parent = c.continuations.top();
    continuations.push(new GRCNode*[size]);
    GRCNode **child = continuations.top();
    for ( int i = 0 ; i < size ; i++ ) child[i] = parent[i];
  }

  void push() { push(*this); }

  void pop() {
    delete [] continuations.top();
    continuations.pop();
  }

  GRCNode *& operator ()(int k) { return continuations.top()[k]; }
};
#line 1138 "ASTGRC.nw"
  
#line 146 "ASTGRC.nw"
class GrcWalker : public Visitor {
protected:
  Context &context;
  GrcSynth &environment;
public:
  GrcWalker(Context &c, GrcSynth &e) : context(c), environment(e) {};

  GRCNode *synthesize(ASTNode *n) {
    assert(n);
    n->welcome(*this);
    assert(context(0));
    return context(0);
  }

  GRCNode *recurse(ASTNode *n) {
    context.push();
    GRCNode *nn = synthesize(n);
    context.pop();
    return nn;
  }

  static GRCNode* push_onto(GRCNode *& b, GRCNode* n) {
    *n >> b;
    b = n;
    return b;
  }
};
#line 1139 "ASTGRC.nw"
  
#line 176 "ASTGRC.nw"
class Surface : public GrcWalker {
public:
  Surface(Context &c, GrcSynth &e) : GrcWalker(c, e) {}
  
#line 204 "ASTGRC.nw"
Status visit(Pause &);
#line 262 "ASTGRC.nw"
Status visit(Exit &s) {
  assert(s.trap);
  assert(s.trap->code > 0);
  assert(context(s.trap->code));
  context(0) = context(s.trap->code);
  return Status();
}
#line 287 "ASTGRC.nw"
Status visit(Emit &s) {
  Action *a = new Action(&s);
  *a >> context(0);
  context(0) = a;
  return Status();
}
#line 309 "ASTGRC.nw"
Status visit(Assignment &s) {
  Action *a = new Action(&s);
  *a >> context(0);
  context(0) = a;
  return Status();
}
#line 342 "ASTGRC.nw"
Status visit(IfThenElse &);
#line 386 "ASTGRC.nw"
Status visit(StatementList &);
#line 468 "ASTGRC.nw"
Status visit(Loop &);
#line 515 "ASTGRC.nw"
Status visit(Suspend &);
#line 628 "ASTGRC.nw"
Status visit(Abort &);
#line 841 "ASTGRC.nw"
Status visit(ParallelStatementList &);
#line 1017 "ASTGRC.nw"
Status visit(Trap &);
#line 1086 "ASTGRC.nw"
Status visit(Signal &);
#line 180 "ASTGRC.nw"
};
#line 1140 "ASTGRC.nw"
  
#line 184 "ASTGRC.nw"
class Depth : public GrcWalker {
public:
  Depth(Context &c, GrcSynth &e) : GrcWalker(c, e) {}
  
#line 222 "ASTGRC.nw"
Status visit(Pause &);
#line 272 "ASTGRC.nw"
Status visit(Exit &) { return Status(); }
#line 296 "ASTGRC.nw"
Status visit(Emit &) { return Status(); }
#line 318 "ASTGRC.nw"
Status visit(Assignment &) { return Status(); }
#line 364 "ASTGRC.nw"
Status visit(IfThenElse &);
#line 390 "ASTGRC.nw"
Status visit(StatementList &);
#line 472 "ASTGRC.nw"
Status visit(Loop &);
#line 519 "ASTGRC.nw"
Status visit(Suspend &);
#line 632 "ASTGRC.nw"
Status visit(Abort &);
#line 845 "ASTGRC.nw"
Status visit(ParallelStatementList &);
#line 1021 "ASTGRC.nw"
Status visit(Trap &);
#line 1090 "ASTGRC.nw"
Status visit(Signal &);
#line 188 "ASTGRC.nw"
};
#line 1141 "ASTGRC.nw"
  
#line 192 "ASTGRC.nw"
class SelTree : public GrcWalker {
public:
  SelTree(Context &c, GrcSynth &e): GrcWalker(c,e) {}
  
#line 241 "ASTGRC.nw"
Status visit(Pause &);
#line 276 "ASTGRC.nw"
Status visit(Exit &) {
  context(0)=new STref();
  return Status();
}
#line 300 "ASTGRC.nw"
Status visit(Emit &) {
  context(0)=new STref();
  return Status();
}
#line 324 "ASTGRC.nw"
Status visit(IfThenElse &);
#line 394 "ASTGRC.nw"
Status  visit(StatementList &s);
#line 476 "ASTGRC.nw"
Status visit(Loop &s);
#line 523 "ASTGRC.nw"
Status visit(Suspend &);
#line 636 "ASTGRC.nw"
Status visit(Abort &);
#line 849 "ASTGRC.nw"
Status visit(ParallelStatementList &);
#line 1025 "ASTGRC.nw"
Status visit(Trap &);
#line 1094 "ASTGRC.nw"
Status visit(Signal &);
#line 196 "ASTGRC.nw"
};
#line 1142 "ASTGRC.nw"
  
#line 55 "ASTGRC.nw"
struct GrcSynth {
  Module *module;
  Context surface_context;
  Context depth_context;
  Context seltree_context;
  Surface surface;
  Depth depth;
  SelTree seltree;
  map<GRCNode *, STNode *> grc2st;

  BuiltinTypeSymbol *integer_type;
  BuiltinTypeSymbol *boolean_type; 

  
#line 73 "ASTGRC.nw"
GrcSynth(Module *m)
  : module(m),
    surface_context(m->max_code+1), depth_context(m->max_code+1), seltree_context(m->max_code+1),
    surface(surface_context, *this), depth(depth_context, *this), seltree(seltree_context, *this)
{
  assert(m->types);
  integer_type = dynamic_cast<BuiltinTypeSymbol*>(m->types->get("integer"));
  assert(integer_type);
  boolean_type = dynamic_cast<BuiltinTypeSymbol*>(m->types->get("boolean"));
  assert(boolean_type);
}
#line 87 "ASTGRC.nw"
GRCNode *synthesize()
{
  GRCNode *synt_seltree, *synt_surface, *synt_depth;

  TopGRC *top;
  Switch *top_switch;
  Leave *lv_flow, *lv_boot;
  STexcl *stroot;
  STleaf *boot, *finished;
  Enter *enfinished, *finalloop;
  Terminate *term0, *term1;

  assert(module->body);

   stroot=new STexcl();
   boot=new STleaf();
   finished=new STleaf();
   finished->setfinal();

   top=new TopGRC();
   lv_boot =new Leave(); lv_flow=new Leave();
   enfinished=new Enter();
   finalloop = new Enter();
   top_switch = new Switch();

   lv_boot->st=boot;
   top_switch->st=stroot;
   lv_flow->st=stroot;
   enfinished->st=finished;
   finalloop->st=finished;

   *lv_flow >> enfinished;
   term0=new Terminate(0); *term0 >> lv_flow;
   term1=new Terminate(1);
   surface_context(0) = depth_context(0) = seltree_context(0) = term0;
   surface_context(1) = depth_context(1) = term1;


   synt_seltree=seltree.synthesize(module->body);
   synt_surface=surface.synthesize(module->body);
   synt_depth=depth.synthesize(module->body);

   *stroot >> finished >> synt_seltree >> boot;

   *lv_boot >> synt_surface;
   
   *top_switch >> finalloop >> synt_depth >> lv_boot;

   *top >> stroot >> top_switch;
 
   std::cerr<<"SYNT OK"<<'\n';

    return top;
}
#line 69 "ASTGRC.nw"
};
#line 1143 "ASTGRC.nw"
}
#endif
