#include "cverilog.hpp"
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
#include <map>
#include <vector>

using namespace CVER;

int debug=0;

map<string, wire*> wnames;

wire* wn(string name){

  int i;
  for(i=0;i<(int)name.length();i++)
    if(name[i]=='[' || name[i]==']') name[i]='_';

  if(wnames.count(name) > 0) return wnames[name];
  wnames[name]=new wire(name);
  return wnames[name];
}

vector<wire*> cw; // current inputs to .names
vector<vector<wire*> > wtable; // wtable;
wire *outw; // the wire to be constructed

void clear_w()
{
  cw.clear();
  wtable.clear();
  outw=NULL;
}
void close_w()
{
  orgate *og;
  andgate *ag;
  wire *w;
  int i;

  if(!outw) return;
  if(outw == w_0 || outw == w_1){clear_w(); return;}

  if(wtable.size() == 0) { new strap(outw, w_0); clear_w(); return;}
  if(wtable.size() == 1){
    // single line: and or strap
    if(wtable[0].size() == 0) { new strap(outw, w_1); clear_w(); return;}
    if(wtable[0].size() == 1) { new strap(outw, wtable[0][0]); clear_w(); return;}
    ag = new andgate();
    ag->out = outw; ag->win = wtable[0]; clear_w();
    return;
  }
  // multiple lines: maybe or
  og=new orgate(); og->out = outw;
  for(i=0;i< (int) wtable.size(); i++){
    assert(wtable[i].size() > 0);
    if(wtable[i].size() == 1) w = wtable[i][0]; else {
      if(debug)cerr<<"warning: multiple bits on a line\n";
      ag=new andgate(); w=ag->out=new wire();
      ag->win = wtable[i];
    }
    og->win.push_back(w);
  }
  clear_w();
}

int main(int argc, char *argv[])
{

  char buf[1000];
  wire *w;
  vector<wire*> wline;

  if(argc!=4) std::cerr<<"Usage: ... file.blif file.v file.v.dot\n";

  ifstream bliffile(argv[1]);
  ofstream vfile(argv[2]);
  ofstream dotfile(argv[3]);

  Cverilog cv(vfile, dotfile);

  current->w_prefix="ww_";

  wnames["w_1"]=w_1;
  wnames["w_0"]=w_0;
  wnames["w_finished"]=w_finished;

  int ninputs,i;
  char ch;

  // now parse the blif file

  outw=NULL;
  while(1){
    bliffile.getline(buf,1000); if(bliffile.fail()) break;

    istringstream ss(buf);
    string kw, id, id2;
    ss >> kw;

    if(debug)cerr<<"kw="<<kw<<"*\n";

    if(kw==".model") continue;
    if(kw==".clock") continue;
    if(kw.find(".default")==0) continue;
    if(kw==".end") {close_w(); continue;}

    if(kw==".inputs") {
      while(1){
	ss >> id;
	if(ss.fail()) break;
	w=wn(id); w->input = w->required=1;
	new strap(w, w_0); // no real world signals yet
      }
      continue;
    }

    if(kw==".outputs"){
      while(1){
	ss >> id;
	if(ss.fail()) break;
	w=wn(id); w->monitor = 3; w->output = w->required=1;
      }
      continue;
    }

    if(kw==".names"){
      close_w();
      while(1){
	ss>>id;
	if(ss.fail()) break;
	outw=w=wn(id);
	cw.push_back(w);
      }
      cw.pop_back();
      ninputs = cw.size();
      if(debug)cerr<<"new .names for wire "<<id<<" : "<<ninputs<<" input[s]\n";
      continue;
    }

    if(kw==".latch"){
      close_w();
      d_ff *ff;
      ss >> id;
      ss >> id2;
      ff = new d_ff(wn(id2), wn(id));
      ss>>id; ss>>id; ss>>id;
      if(id=="1") ff->init=1;
      clear_w();
      continue;
    }

    // bit line

    if(debug)cerr<<"  bitline\n";
    wline.clear();
    for(i=0;i<ninputs;i++){
      ch=kw[i];
      if(ch=='-') continue;
      if(ch=='1') wline.push_back(cw[i]);
      if(ch=='0') wline.push_back(current->build_notgate(cw[i]));
    }
    wtable.push_back(wline);

  }

  cv.remove_unused(); // just for bogus wires
  monitor_level = 2; // display only signals
  cv.build();
  cv.drawdot();

  return 0;
}
