#include "cverilog.hpp"

#include <vector>

namespace CVER{

  void Cverilog::build_blif()
  {
    // move w_finished to end, for display purposes
    if(wires[0]==w_finished){
      wires.erase(wires.begin());
      wires.push_back(w_finished);
    }

    o<<".model top\n";
    o<<".inputs";
    for(vector<wire*>::iterator w=wires.begin(); w!=wires.end(); w++)
      if((*w)->input)
	o<<" "<<(*w)->name;
    o<<"\n";

    o<<".outputs";
    for(vector<wire*>::iterator w=wires.begin(); w!=wires.end(); w++)
      if((*w)->output)
	o<<" "<<(*w)->name;
    //    o<<" w_finished";
    o<<"\n";

    o<<".clock clk\n";

    o<<".names w_0\n";
    o<<".names w_1\n"<<" 1\n";

    for(vector<vermod*>::iterator m=modules.begin(); m!=modules.end();m++)
      (*m)->pr_blif();


    o<<".end\n";
  }

  void strap::pr_blif()
  {
    ostream &o=current->o;
    if(out->input) return;
    o<<".names "<<in->name<<" "<<out->name<<"\n";
    o<<"1 1\n";
  }

  void andgate::pr_blif()
  {
    ostream &o=current->o;

    o<<".names";
    for(vector<wire*>::iterator w=win.begin(); w!=win.end(); w++){
      assert(*w);
      o<<" "<<(*w)->name;
    }
    
    assert(out);
    o<<" "<<out->name<<"\n";
    
    for(vector<wire*>::iterator w=win.begin(); w!=win.end(); w++)
      o<<"1";
    o<<" 1\n";
  }

  void orgate::pr_blif()
  {
    ostream &o=current->o;

    o<<".names";
    for(vector<wire*>::iterator w=win.begin(); w!=win.end(); w++){
      assert(*w);
      o<<" "<<(*w)->name;
    }
    assert(out);
    o<<" "<<out->name<<"\n";

    for(vector<wire*>::iterator w=win.begin(); w!=win.end(); w++){
      for(vector<wire*>::iterator w1=win.begin(); w1!=win.end(); w1++)
	o<<((w==w1)?"1":"-");
      o<<" 1\n";
    }

  }
  
  void notgate::pr_blif()
  {
    ostream &o=current->o;
    
    o<<".names";
    assert(in && out);
    o<<" "<<in->name;
    o<<" "<<out->name<<"\n";
    o<<"0 1\n";
  }
  
  void d_ff::pr_blif()
  {
    ostream &o=current->o;
    assert(d && q && rst); // rst not used
    o<<".latch "<<d->name<<" "<<q->name<<" re clk "<<init<<"\n";

  }
  
  void counter::pr_blif()
  {
    int i, j, sz, t;
    ostream &o=current->o;
    
    vector <string> q,d,m1;
    vector <int> init;
    string cnor, all0;

    sz=0; t=1;
    while(t<=startval){sz++; t+=t;}
    cerr<<"... using "<<sz<<" bits for startval="<<startval<<"\n";

    q.resize(sz); d.resize(sz); m1.resize(sz); init.resize(sz);

    for(i=0;i<sz;i++){
      init[i] = startval & 1;
      startval >>= 1;
    }

    for(i=0;i<sz;i++){
      d[i]=current->new_wname();
      q[i]=current->new_wname();
      m1[i]=current->new_wname();
      o<<".latch "<<d[i]<<" "<<q[i]<<" re clk 0\n";
      
    }
    
    all0 = current->new_wname();

    o<<".names"; for(i=0;i<sz;i++)o<<" "<<q[i]; o<<" "<<all0<<"\n";
    for(i=0;i<sz;i++)o<<"0"; o<<" 1\n";

    for(i=0;i<sz;i++){
      cnor = current->new_wname();
      o<<".names";
      for(j=0;j<i;j++) o<<" "<<q[j];
      o<<" "<<cnor<<"\n";
      for(j=0;j<i;j++) o<<"0";
      o<<" "<<"1\n";

      o<<".names "<<q[i]<<" "<<cnor<<" "<<m1[i]<<"\n";
      o<<"01 1\n";
      o<<"10 1\n";

      o<<".names "<<start->name<<" "<<all0<<" "<<cenable->name<<" "<<q[i]<<" "<<m1[i]<<" "<<d[i]<<"\n";
      if(init[i])o<<"1---- 1\n"; // init
      o<<"001-1 1\n"; // not init, not =0=, cenable=1 -> q-1
      o<<"0001- 1\n"; //      -  "  -       cenable=0 -> q

    }

    o<<".names "<<all0<<" "<<cenable->name<<" "<<alarm->name<<"\n";
    o<<"11 1\n";

  }

}
