#!/bin/sh

# Simple testing framework for running the GRC dismantler

STRLXML=./strlxml
DISMANTLE=./cec-dismantle
ASTGRC=./cec-astgrc
GRCDOT=./grcdot
DOT=dot
GRCOPT=./grcopt

globallog=testgrc.log
rm -f $globallog
error=0

Check() {
    basename=`echo $1 | sed 's/.*\\///
                             s/.strl//'`
    xmlfile=${basename}.xml
    disfile=${basename}.dis
    grcfile=${basename}.grc
    grcoptfile=${basename}.opt.grc
    dotfile=${basename}.dot
    dotoptfile=${basename}.opt.dot
    psfile=${basename}.ps
    psoptfile=${basename}.opt.ps

    if [ "x$2" != "xnopdf" -a "x$3" != "xnopdf" ]; then
	rundot=1
    fi


    echo -n "Generating GRC for $basename..."
    echo "Generating GRC for $basename" 1>&2
    $STRLXML < $1 > $xmlfile || {
	echo "FAILED: strlxml terminated"
        error=1
	return 1
    }
    $DISMANTLE < $xmlfile > $disfile || {
	echo "FAILED: cec-dismantle terminated"
        error=1
	return 1
    }
    $ASTGRC < $disfile > $grcfile || {
	echo "FAILED: cec-astgrc terminated"
        error=1
	return 1
    }
    echo OK

    if [ "x$rundot" = "x1" -a "x$2" != "xopt" ]; then

    echo "Running dot for nonoptimized GRC ..."
    $GRCDOT < $grcfile > $dotfile || {
	echo "FAILED: grcdot terminated"
        error=1
	return 1
    }
    $DOT -Tps < $dotfile > $psfile || {
	echo "FAILED: dot terminated"
        error=1
	return 1
    }
    else
	rm -f $psfile
    fi

# grcopt: rewrites the dot to match labels

    if [ "x$2" = "xopt" ]; then

    echo -n "Running grcopt ..."
    $GRCOPT $grcfile $grcoptfile $dotfile $dotoptfile || {
	echo "FAILED: grcopt terminated"
        error=1
	return 1
    }
    echo "OK"

    if [ "x$rundot" = "x1" ]; then

    echo "Running dot for {non,}optimized GRC ..."
    cat $dotfile $1 | ./addstrl2dot | $DOT -Tps > $psfile || {
	echo "FAILED: dot terminated"
        error=1
	return 1
    }

    cat $dotoptfile $1 | ./addstrl2dot | $DOT -Tps > $psoptfile || {
	echo "FAILED: dot terminated for opt"
        error=1
	return 1
    }
    # rm $xmlfile $disfile $outfile $grcfile
    else
	rm -f $psfile $psoptfile
    fi

    fi

}

#for file in tests/test*.strl
#do
#  Check $file 2>> $globallog
#done

Check $1 $2 $3 2>$globallog

exit $error
